/******************************************************************************/
/* (c) 2006 jamie twycross, jpt AT cs.nott.ac.uk                              */
/* released under gnu gpl v2                                                  */
/* tcuna - systrace tissue client user notification agent                     */
/******************************************************************************/

#include <stdio.h>
#include <sys/poll.h>
#include <err.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <unistd.h>
#include <signal.h>

#include <tissue/tissue.h>
#include <tissue/syscall.h>

#define TIMEOUT			1000
#define MAX_LINE		4096
#define OUTFILE			"tcuna1.log"
#define CS_DEFAULT_PORT	5049
#define BACKLOG			10

static char tshostname[256] = "localhost";
static int16_t tsport = TS_DEFAULT_PORT;
static char clientname[256] = "tcuna1";  /* name of client */
static char logfile[64] = "tcuna1.log";  /* logfile */

int remove_policy(char *command)
{
	char policy_file[MAX_LINE], policy[MAX_LINE], tmp_file[MAX_LINE];
	char line[MAX_LINE], *p;
	FILE *ifp, *ofp;
	int pid;

	/* parse command, format: "pid syscall-name: filter" */
	pid = atoi(command);
	if(get_policy_filename(pid, policy_file))
		return 1;  /* invalid pid */
	snprintf(tmp_file, MAX_LINE, "/tmp/tcuna.%d.tmp", rand());
	if((p = index(command, ' ')) == NULL)
		return 1;  /* ignore malformed commands */
	strncpy(policy, p + 1, MAX_LINE);

	/* create new policy file in temp file */
	if((ifp = fopen(policy_file, "r")) == NULL)
		return 1;
	if((ofp = fopen(tmp_file, "w")) == NULL) {
		fclose(ifp);
		return 1;
	}
	while(freadline(line, MAX_LINE, ifp) != NULL)
		if(strstr(line, policy) == NULL)
			fprintf(ofp, "%s\n", line);
	fclose(ifp);
	fclose(ofp);

	/* copy over old policy */
	if((ifp = fopen(tmp_file, "r")) == NULL)
		return 1;
	if((ofp = fopen(policy_file, "w")) == NULL) {
		fclose(ifp);
		return 1;
	}
	while(freadline(line, MAX_LINE, ifp) != NULL)
		fprintf(ofp, "%s\n", line);
	fclose(ifp);
	fclose(ofp);
	remove(tmp_file);

	return 0;
}

int add_policy(char *command)
{
	char policy_file[MAX_LINE], policy[MAX_LINE], tmp_file[MAX_LINE];
	char syscall[MAX_LINE];
	char line[MAX_LINE], *p1, *p2;
	int pid;
	FILE *ifp, *ofp;

	/* parse command, format: "pid syscall-name: filter" */
	pid = atoi(command);
	if(get_policy_filename(pid, policy_file))
		return 1;  /* invalid pid */
	snprintf(tmp_file, MAX_LINE, "/tmp/tcuna.%d.tmp", rand());
	if((p1 = index(command, ' ')) == NULL)
		return 1;  /* ignore malformed commands */
	strncpy(policy, p1 + 1, MAX_LINE);
	if(((p2 = index(command, ':')) == NULL) || (p1 > p2))
		return 1;  /* ignore malformed commands */
	memcpy(syscall, p1 + 1, (p2 - p1 - 2) * sizeof(char));
	syscall[p2 - p1 - 2] = '\0';

	/* create new policy file in temp file */
	if((ifp = fopen(policy_file, "r")) == NULL)
		return 1;
	if((ofp = fopen(tmp_file, "w")) == NULL) {
		fclose(ifp);
		return 1;
	}
	while(freadline(line, MAX_LINE, ifp) != NULL)
		if(strstr(line, syscall) == NULL)
			fprintf(ofp, "%s\n", line);
	fprintf(ofp, "\t%s\n", policy);
	fclose(ifp);
	fclose(ofp);

	/* copy over old policy */
	if((ifp = fopen(tmp_file, "r")) == NULL)
		return 1;
	if((ofp = fopen(policy_file, "w")) == NULL) {
		fclose(ifp);
		return 1;
	}
	while(freadline(line, MAX_LINE, ifp) != NULL)
		fprintf(ofp, "%s\n", line);
	fclose(ifp);
	fclose(ofp);
	remove(tmp_file);

	return 0;
}

void start_command_server(unsigned short int port, FILE *fp)
{
	struct sockaddr_in sa;
	FILE *nfp;
	int sd, nsd;
	socklen_t sas;
	char line[MAX_LINE];
	time_t curtime;

	/* fork server */
	switch(fork()) {
		/* child */
		case 0:
			/* create socket */
			sd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
			if(sd == -1)
				err(1, "tcuna command server error");

			/* connect socket to port */
			sa.sin_family = AF_INET;
			sa.sin_addr.s_addr = INADDR_ANY;
			sa.sin_port = htons(port);
			memset(&sa.sin_zero, 0, 8);
			if(bind(sd, (struct sockaddr *) &sa, sizeof(struct sockaddr)) == -1)
				err(1, "tcuna command server error");
			if(listen(sd, BACKLOG) == -1)
				err(1, "tcuna command server error");

			/* just echo input to stdout i.e. systrace */
			while(1) {
				sas = sizeof(struct sockaddr_in);
				if((nsd = accept(sd, (struct sockaddr *) &sa, &sas)) == -1)
					err(1, "tcuna command server error");
				if(!(nfp = fdopen(nsd, "r")))
					err(1, "tcuna command server error");
				if(freadline(line, MAX_LINE, nfp) == NULL)
					err(1, "tcuna command server read error");

				/* parse input */
				if(strstr(line, "rmpolicy ") == line) {
					/* remove policy */
					curtime = time(NULL);
					fprintf(fp, "tcuna: %s %s\n", strtok(ctime(&curtime), \
							"\n"), line);
					if(remove_policy(line + 9) == 0)
						fprintf(fp, "tcuna: OKAY rmpolicy\n");
					else
						fprintf(fp, "tcuna: WRONG rmpolicy\n");
					fflush(fp);
				} else {
					if(strstr(line, "addpolicy ") == line) {
						/* add policy */
						curtime = time(NULL);
						fprintf(fp, "tcuna: %s %s\n", strtok(ctime(&curtime), \
								"\n"), line);
						if(add_policy(line + 10) == 0)
							fprintf(fp, "tcuna: OKAY addpolicy\n");
						else
							fprintf(fp, "tcuna: WRONG addpolicy\n");
						fflush(fp);
					} else {
						/* pass message along to systrace */
						printf("%s\n", line);
						fflush(stdout);
						curtime = time(NULL);
						fprintf(fp, "tcuna: %s %s\n", strtok(ctime(&curtime), \
								"\n"), line);
						fflush(fp);
					}
				}

				/* tidy up */
				fclose(nfp);
				close(nsd);
			}
			break;

		/* parent */
		default:
			break;
	}
}

/* convert systrace request to datastream structure */
DS *streq2ds(Systrace_Request *request, DS *ds)
{
	char line[MAX_LINE], *p1, *p2;

	/* set id to pid */
	ds->id = atoi(request->process_id);

	/* set timestamp to time of request */
	memcpy(&ds->tm, &request->tm, sizeof(struct timeval));

	/* set first antigen to syscall number + number of args */
	if((p1 = index(request->syscall_info, '(')) == NULL)
		return NULL;
	if((p2 = index(request->syscall_info, ':')) == NULL)
		return NULL;
	snprintf(line, MAX_LINE, "%d-%d", atoi(p1 + 1), atoi(p2 + 1));
	ds->antigen[0] = set_antigen(ds->antigen[0], line);
	ds->antigen[0]->id = ds->id;  /* same id as datastream */

	return ds;
}

static void sigintterm_handler(int s)
{
	/* tidy up */
	close_log();
	disconnect_ts();
	_exit(EXIT_SUCCESS);
}

int main(int argc, char **argv)
{
	struct pollfd pfd;
	int ret;
	char line[MAX_LINE];
	Systrace_Request request;
	DS *ds;
	Network_Info ts_net_info; /* network info on tissue server */
	char message[MAX_LINE];

	/* connect to tissue server */
	strcpy(ts_net_info.hostname, tshostname);
	ts_net_info.port = (int16_t) tsport;
	strcpy(ts_net_info.clientname, clientname);
	connect_ts(&ts_net_info);

	/* capture SIGINT + SIGTERM */
	if(signal(SIGINT, sigintterm_handler) == SIG_ERR || \
			signal(SIGTERM, sigintterm_handler) == SIG_ERR) {
		err(1, "couldn't setup signal handling");
		_exit(EXIT_FAILURE);
	}

	/* initialise antigen datastream structure */
	ds = init_ds(0, 1);   /* 0 signals, 1 antigen = syscall number */

	/* open log file */
	open_log(logfile);

	/* start command server */
/*	start_command_server(CS_DEFAULT_PORT, fp); */

	/* wait for input messages from systrace */
	setvbuf(stdin, NULL, _IONBF, 0);
	setvbuf(stdout, NULL, _IONBF, 0);
	while(1) {
		/* wait for input on stdin */
		memset(&pfd, 0, sizeof(pfd));
		pfd.fd = fileno(stdin);
		pfd.events = POLLIN;
		ret = poll(&pfd, 1, TIMEOUT);
		if(ret == -1)
			err(1, NULL);
		if(ret == 0)
			continue;

		/* read + parse input */
		if(freadline(line, MAX_LINE, stdin) == NULL) {
			if(feof(stdin))
				continue;
			else
				err(1, "tcuna read error");
		}

		/* write to log */
		snprintf(message, sizeof(message) - 1, "systrace %s", line);
		write_log(message);

		if(systrace_request(line, &request) == 0) {
			/* always permit-now systrace request first */
			printf("permit-now\n");
			fflush(stdout);
			/* process request */
			if(streq2ds(&request, ds))
				send_ds(ds);
		}
	}

	return 0;
}
